/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import QtQuick 2.0
import Ubuntu.Components 0.1 as Ubuntu

/*
  This component is a unified text selection and scrolling handler for both
  TextField and TextArea components.
  */

Item {
    id: inputHandler
    objectName: "input_handler"
    // the root control
    property Item main
    // the input instance
    property Item input
    // the Flickable holdiong the input instance
    property Flickable flickable
    // selection cursor mode
    property bool selectionCursor: input && input.selectedText !== ""
    // True if mouse handlig is enabled, false if flicking mode is enabled
    readonly property bool mouseHandlingEnabled: !flickable.interactive
    // property holding the selection mode timeout
    property int selectionModeTimeout: 200

    // signal triggered when popup shoudl be opened
    signal pressAndHold(int pos)

    function activateInput() {
        if (!input.activeFocus) {
            input.forceActiveFocus();
        } else {
            showInputPanel();
        }
    }

    function showInputPanel() {
        if (!Qt.inputMethod.visible) {
            Qt.inputMethod.show();
        }
        textChanged = false;
    }
    function hideInputPanel() {
        Qt.inputMethod.hide();
        // emit accepted signal if changed
        if (textChanged && input.hasOwnProperty("accepted")) {
            input.accepted();
        }
    }

    // internal properties/functions
    readonly property bool singleLine: input && input.hasOwnProperty("validator")
    property var flickableList: new Array()
    property bool textChanged: false
    property int pressedPosition: -1
    // move properties
    property int moveStarts: -1
    property int moveEnds: -1
    // set scroller to the first Flickable that scrolls the input
    // this can be the internal Flickable if the full autosize is disabled
    // or one of the input's parent Flickable
    readonly property bool scrollingDisabled: main && main.hasOwnProperty("autoSize") ?
                                                  (main.autoSize && (main.maximumLineCount <= 0)) : false
    onScrollingDisabledChanged: if (state == "") flickable.interactive = !scrollingDisabled
    readonly property Flickable grandScroller: firstFlickableParent(main)
    readonly property Flickable scroller: (scrollingDisabled && grandScroller) ? grandScroller : flickable

    // ensures the text cusrorRectangle is always in the internal Flickable's visible area
    function ensureVisible()
    {
        var rect = input.cursorRectangle;
        if (flickable.moving || flickable.flicking)
            return;
        if (flickable.contentX >= rect.x)
            flickable.contentX = rect.x;
        else if (flickable.contentX + flickable.width <= rect.x + rect.width)
            flickable.contentX = rect.x + rect.width - flickable.width;
        if (flickable.contentY >= rect.y)
            flickable.contentY = rect.y;
        else if (flickable.contentY + flickable.height <= rect.y + rect.height)
            flickable.contentY = rect.y + rect.height - flickable.height;
    }
    // returns the mouse position
    function mousePosition(mouse) {
        return singleLine ? input.positionAt(mouse.x) : input.positionAt(mouse.x, mouse.y);
    }
    // checks whether the position is in the selected text
    function positionInSelection(pos) {
        return (input.selectionStart !== input.selectionEnd)
                && (pos >= Math.min(input.selectionStart, input.selectionEnd))
                && (pos <= Math.max(input.selectionStart, input.selectionEnd));
    }

    // check whether the mouse is inside a selected text area
    function mouseInSelection(mouse) {
        var pos = mousePosition(mouse);
        return positionInSelection(pos);
    }
    // selects text
    function selectText(mouse) {
        moveEnds = mousePosition(mouse);
        if (moveStarts < 0) {
            moveStarts = moveEnds;
        }
        input.select(moveStarts, moveEnds);
    }
    // returns the first Flickable parent of a given item
    function firstFlickableParent(item) {
        var p = item ? item.parent : null;
        while (p && !p.hasOwnProperty("flicking")) {
            p = p.parent;
        }
        return p;
    }

    // disables interactive Flickable parents, stops at the first non-interactive flickable.
    function toggleFlickablesInteractive(turnOn) {
        var p;
        if (!turnOn) {
            // handle the scroller separately
            p = firstFlickableParent(scroller)
            while (p) {
                if (p.interactive) {
                    flickableList.push(p);
                    p.interactive = false;
                } else {
                    break;
                }
                p = firstFlickableParent(p);
            }
        } else {
            while (flickableList.length > 0) {
                p = flickableList.pop();
                p.interactive = true;
            }
        }
    }

    Component.onCompleted: {
        state = (main.focus) ? "" : "inactive";
    }

    // states
    states: [
        // override default state to turn on the saved Flickable interactive mode
        State {
            name: ""
            StateChangeScript {
                // restore interactive for all Flickable parents
                script: toggleFlickablesInteractive(true);
            }
        },
        State {
            name: "inactive"
            // we do not disable scroller here as in case the internal scrolling
            // is disabled (scrollingDisabled = true) the outer scroller (grandScroller)
            // would be blocked as well, which we don't want to
            PropertyChanges {
                target: flickable
                interactive: false
            }
        },
        State {
            name: "scrolling"
            StateChangeScript {
                script: {
                    // stop scrolling all the parents
                    toggleFlickablesInteractive(false);
                    // stop selection timeout
                    selectionTimeout.running = false;
                }
            }
        },
        State {
            name: "select"
            // during select state all the flickables are blocked (interactive = false)
            // we can use scroller here as we need to disable the outer scroller too!
            PropertyChanges {
                target: scroller
                interactive: false
            }
            StateChangeScript {
                script: {
                    // turn off interactive for all parent flickables
                    toggleFlickablesInteractive(false);
                    if (!positionInSelection(pressedPosition)) {
                        input.cursorPosition = pressedPosition;
                    }
                }
            }
        }
    ]

    // brings the state back to default when the component looses focuse
    Connections {
        target: main
        ignoreUnknownSignals: true
        onFocusChanged: {
            state = (main.focus) ? "" : "inactive";
        }
    }

    // input specific signals
    Connections {
        target: input
        onCursorRectangleChanged: if (!scrollingDisabled) ensureVisible()
        onTextChanged: textChanged = true;
    }

    // inner or outer Flickable controlling
    Connections {
        target: scroller
        // turn scrolling state on
        onFlickStarted: if (!scrollingDisabled) state = "scrolling"
        onMovementStarted: if (!scrollingDisabled) state = "scrolling"
        // reset to default state
        onMovementEnded: state = ""
    }

    // switches the state to selection
    Timer {
        id: selectionTimeout
        interval: selectionModeTimeout
        onTriggered: {
            if (scroller && !scroller.moving) {
                state = "select";
            }
        }
    }

    // Mouse handling
    Ubuntu.Mouse.forwardTo: [main]
    Ubuntu.Mouse.clickAndHoldThreshold: units.gu(2)
    Ubuntu.Mouse.onPressed: {
        if (input.activeFocus) {
            // start selection timeout
            selectionTimeout.restart();
        }
        // remember pressed position as we need it when entering into selection state
        pressedPosition = mousePosition(mouse);
        // consume event so it does not get forwarded to the input
        mouse.accepted = true;
    }
    Ubuntu.Mouse.onReleased: {
        if (!main.focus && !main.activeFocusOnPress) {
            return;
        }

        activateInput();
        // stop text selection timer
        selectionTimeout.running = false;
        if (state === "") {
            input.cursorPosition = mousePosition(mouse);
        }
        moveStarts = moveEnds = -1;
        state = "";
    }
    Ubuntu.Mouse.onPositionChanged: {
        // leave if not focus, not the left button or not in select state
        if (!input.activeFocus || (mouse.button !== Qt.LeftButton) || (state !== "select") || !main.selectByMouse) {
            return;
        }
        // stop text selection timer
        selectionTimeout.running = false;
        selectText(mouse);
    }
    Ubuntu.Mouse.onDoubleClicked: {
        if (main.selectByMouse) {
            input.selectWord();
            // turn selection state temporarily so the selection is not cleared on release
            state = "selection";
        }
    }
    Ubuntu.Mouse.onPressAndHold: {
        if ((state === "select") && mouseInSelection(mouse)) {
            // trigger clipboard popover
            pressAndHold(input.cursorPosition);
            // get out of the select mode, otherwise we mey change the selection by dragging the mouse while pressed
            state = "";
        }
    }
}
